using System;
using System.Collections.Generic;
using System.Text;
using Zcu.Mve.Core;
using Zcu.Mve.Visualization;
using Zcu.Mve.Numerics;

namespace Examples
{
	class AnimationDistanceEvaluatorFast:Module
	{
		UnstrGrid result = null;

		public AnimationDistanceEvaluatorFast()
		{
			AddInPort("Evaluated Mesh", typeof(UnstrGrid));
			AddInPort("Target Mesh", typeof(UnstrGrid));
			AddOutPort("Result", typeof(UnstrGrid));
		}

		private List<int>[, , ,] pointsList;
		private List<int>[, , ,] edgesList;
		private List<int>[, , ,] facesList;
		private List<int>[, , ,] tetrahedraList;
		private Line[] edges;
		private Triangle[] faces;
		private UniformDataArray points;
		private UniformDataArray tetrahedra;
		private double[] mins, steps;
		private int[] divs;
		private int[] lineChecks, faceChecks, tetraChecks;

		private void TetrahedronBoundingBox(Tetrahedron t, UniformDataArray points, out double[] mins, out double[] maxs)
		{
			mins = new double[] { double.PositiveInfinity, double.PositiveInfinity, double.PositiveInfinity, double.PositiveInfinity };
			maxs = new double[] { double.NegativeInfinity, double.NegativeInfinity, double.NegativeInfinity, double.NegativeInfinity };

			PointND p = (PointND)points[t.V1];
			if (p[0] < mins[0]) mins[0] = p[0];
			if (p[1] < mins[1]) mins[1] = p[1];
			if (p[2] < mins[2]) mins[2] = p[2];
			if (p[3] < mins[3]) mins[3] = p[3];
			if (p[0] > maxs[0]) maxs[0] = p[0];
			if (p[1] > maxs[1]) maxs[1] = p[1];
			if (p[2] > maxs[2]) maxs[2] = p[2];
			if (p[3] > maxs[3]) maxs[3] = p[3];

			p = (PointND)points[t.V2];
			if (p[0] < mins[0]) mins[0] = p[0];
			if (p[1] < mins[1]) mins[1] = p[1];
			if (p[2] < mins[2]) mins[2] = p[2];
			if (p[3] < mins[3]) mins[3] = p[3];
			if (p[0] > maxs[0]) maxs[0] = p[0];
			if (p[1] > maxs[1]) maxs[1] = p[1];
			if (p[2] > maxs[2]) maxs[2] = p[2];
			if (p[3] > maxs[3]) maxs[3] = p[3];

			p = (PointND)points[t.V3];
			if (p[0] < mins[0]) mins[0] = p[0];
			if (p[1] < mins[1]) mins[1] = p[1];
			if (p[2] < mins[2]) mins[2] = p[2];
			if (p[3] < mins[3]) mins[3] = p[3];
			if (p[0] > maxs[0]) maxs[0] = p[0];
			if (p[1] > maxs[1]) maxs[1] = p[1];
			if (p[2] > maxs[2]) maxs[2] = p[2];
			if (p[3] > maxs[3]) maxs[3] = p[3];

			p = (PointND)points[t.V4];
			if (p[0] < mins[0]) mins[0] = p[0];
			if (p[1] < mins[1]) mins[1] = p[1];
			if (p[2] < mins[2]) mins[2] = p[2];
			if (p[3] < mins[3]) mins[3] = p[3];
			if (p[0] > maxs[0]) maxs[0] = p[0];
			if (p[1] > maxs[1]) maxs[1] = p[1];
			if (p[2] > maxs[2]) maxs[2] = p[2];
			if (p[3] > maxs[3]) maxs[3] = p[3];
		}

		private void TriangleBoundingBox(Triangle t, UniformDataArray points, out double[] mins, out double[] maxs)
		{
			mins = new double[] { double.PositiveInfinity, double.PositiveInfinity, double.PositiveInfinity, double.PositiveInfinity };
			maxs = new double[] { double.NegativeInfinity, double.NegativeInfinity, double.NegativeInfinity, double.NegativeInfinity };

			PointND p = (PointND)points[t.V1];
			if (p[0] < mins[0]) mins[0] = p[0];
			if (p[1] < mins[1]) mins[1] = p[1];
			if (p[2] < mins[2]) mins[2] = p[2];
			if (p[3] < mins[3]) mins[3] = p[3];
			if (p[0] > maxs[0]) maxs[0] = p[0];
			if (p[1] > maxs[1]) maxs[1] = p[1];
			if (p[2] > maxs[2]) maxs[2] = p[2];
			if (p[3] > maxs[3]) maxs[3] = p[3];

			p = (PointND)points[t.V2];
			if (p[0] < mins[0]) mins[0] = p[0];
			if (p[1] < mins[1]) mins[1] = p[1];
			if (p[2] < mins[2]) mins[2] = p[2];
			if (p[3] < mins[3]) mins[3] = p[3];
			if (p[0] > maxs[0]) maxs[0] = p[0];
			if (p[1] > maxs[1]) maxs[1] = p[1];
			if (p[2] > maxs[2]) maxs[2] = p[2];
			if (p[3] > maxs[3]) maxs[3] = p[3];

			p = (PointND)points[t.V3];
			if (p[0] < mins[0]) mins[0] = p[0];
			if (p[1] < mins[1]) mins[1] = p[1];
			if (p[2] < mins[2]) mins[2] = p[2];
			if (p[3] < mins[3]) mins[3] = p[3];
			if (p[0] > maxs[0]) maxs[0] = p[0];
			if (p[1] > maxs[1]) maxs[1] = p[1];
			if (p[2] > maxs[2]) maxs[2] = p[2];
			if (p[3] > maxs[3]) maxs[3] = p[3];			
		}

		private void LineBoundingBox(Line l, UniformDataArray points, out double[] mins, out double[] maxs)
		{
			mins = new double[] { double.PositiveInfinity, double.PositiveInfinity, double.PositiveInfinity, double.PositiveInfinity };
			maxs = new double[] { double.NegativeInfinity, double.NegativeInfinity, double.NegativeInfinity, double.NegativeInfinity };

			PointND p = (PointND)points[l.V1];
			if (p[0] < mins[0]) mins[0] = p[0];
			if (p[1] < mins[1]) mins[1] = p[1];
			if (p[2] < mins[2]) mins[2] = p[2];
			if (p[3] < mins[3]) mins[3] = p[3];
			if (p[0] > maxs[0]) maxs[0] = p[0];
			if (p[1] > maxs[1]) maxs[1] = p[1];
			if (p[2] > maxs[2]) maxs[2] = p[2];
			if (p[3] > maxs[3]) maxs[3] = p[3];

			p = (PointND)points[l.V2];
			if (p[0] < mins[0]) mins[0] = p[0];
			if (p[1] < mins[1]) mins[1] = p[1];
			if (p[2] < mins[2]) mins[2] = p[2];
			if (p[3] < mins[3]) mins[3] = p[3];
			if (p[0] > maxs[0]) maxs[0] = p[0];
			if (p[1] > maxs[1]) maxs[1] = p[1];
			if (p[2] > maxs[2]) maxs[2] = p[2];
			if (p[3] > maxs[3]) maxs[3] = p[3];			
		}

		private void addLine(int index)
		{
			Line l = edges[index];
			double[] tmins, tmaxs;
			LineBoundingBox(l, points, out tmins, out tmaxs);

			int[] minI;
			int[] maxI;
			FindRange(tmins, tmaxs, out minI, out maxI);

			// add tetrahedron to given slots
			for (int xi = minI[0]; xi <= maxI[0]; xi++)
				for (int yi = minI[1]; yi <= maxI[1]; yi++)
					for (int zi = minI[2]; zi <= maxI[2]; zi++)
						for (int ti = minI[3]; ti <= maxI[3]; ti++)
							edgesList[xi, yi, zi, ti].Add(index);
		}

		private void addTriangle(int index)
		{
			Triangle t = faces[index];
			double[] tmins, tmaxs;
			TriangleBoundingBox(t, points, out tmins, out tmaxs);

			int[] minI;
			int[] maxI;
			FindRange(tmins, tmaxs, out minI, out maxI);

			// add tetrahedron to given slots
			for (int xi = minI[0]; xi <= maxI[0]; xi++)
				for (int yi = minI[1]; yi <= maxI[1]; yi++)
					for (int zi = minI[2]; zi <= maxI[2]; zi++)
						for (int ti = minI[3]; ti <= maxI[3]; ti++)
							facesList[xi, yi, zi, ti].Add(index);
		}

		private void FindRange(double[] tmins, double[] tmaxs, out int[] minI, out int[] maxI)
		{
			minI = new int[] { 0, 0, 0, 0 };
			maxI = new int[] { divs[0], divs[1], divs[2], divs[3] };

			// find X span
			double x = mins[0] + steps[0];
			while (tmins[0] > x)
			{
				minI[0]++;
				x += steps[0];
			}
			maxI[0] = minI[0];
			while (tmaxs[0] > x)
			{
				maxI[0]++;
				x += steps[0];
			}

			// find Y span
			x = mins[1] + steps[1];
			while (tmins[1] > x)
			{
				minI[1]++;
				x += steps[1];
			}
			maxI[1] = minI[1];
			while (tmaxs[1] > x)
			{
				maxI[1]++;
				x += steps[1];
			}

			// find z span
			x = mins[2] + steps[2];
			while (tmins[2] > x)
			{
				minI[2]++;
				x += steps[2];
			}
			maxI[2] = minI[2];
			while (tmaxs[2] > x)
			{
				maxI[2]++;
				x += steps[2];
			}

			// find t span
			x = mins[3] + steps[3];
			while (tmins[3] > x)
			{
				minI[3]++;
				x += steps[3];
			}
			maxI[3] = minI[3];
			while (tmaxs[3] > x)
			{
				maxI[3]++;
				x += steps[3];
			}
}

		private void addTetrahedron(int index)
		{
			Tetrahedron t = (Tetrahedron)tetrahedra[index];
			double[] tmins, tmaxs;
			TetrahedronBoundingBox(t, points, out tmins, out tmaxs);

			int[] minI;
			int[] maxI;
			FindRange(tmins, tmaxs, out minI, out maxI);

			// add tetrahedron to given slots
			for (int xi = minI[0]; xi <= maxI[0]; xi++)
				for (int yi = minI[1]; yi <= maxI[1]; yi++)
					for (int zi = minI[2]; zi <= maxI[2]; zi++)
						for (int ti = minI[3]; ti <= maxI[3]; ti++)
							tetrahedraList[xi, yi, zi, ti].Add(index);
		}

		private void addPoint(int index)
		{
			PointND p = (PointND)points[index];			

			int[] indices = new int[4];
			for (int i = 0; i < 4; i++)
				indices[i] = (int)((p[i] - mins[i]) / steps[i]);

			pointsList[indices[0], indices[1], indices[2], indices[3]].Add(index);
		}



		private void fillFaces(UnstrGrid data)
		{
			Dictionary<Triangle, int> faceHash = new Dictionary<Triangle, int>();			

			foreach (Tetrahedron t in tetrahedra)
			{
				Triangle triangle;

				triangle = new Triangle(t.V1, t.V2, t.V3);
				if (!faceHash.ContainsKey(triangle))
				{
					faceHash.Add(triangle, faceHash.Count);
				}

				triangle = new Triangle(t.V1, t.V2, t.V4);
				if (!faceHash.ContainsKey(triangle))
				{
					faceHash.Add(triangle, faceHash.Count);
				}

				triangle = new Triangle(t.V1, t.V3, t.V4);
				if (!faceHash.ContainsKey(triangle))
				{
					faceHash.Add(triangle, faceHash.Count);
				}

				triangle = new Triangle(t.V2, t.V3, t.V4);
				if (!faceHash.ContainsKey(triangle))
				{
					faceHash.Add(triangle, faceHash.Count);
				}
			}

			faces = new Triangle[faceHash.Count];
			foreach (KeyValuePair<Triangle, int> pair in faceHash)
			{
				faces[pair.Value] = pair.Key;
			}
		}


		private void fillEdges(UnstrGrid data)
		{
			Dictionary<Line, int> edgeHash = new Dictionary<Line, int>();

			UniformDataArray tetrahedra = data.GetCells(typeof(Tetrahedron))[0];

			foreach (Tetrahedron t in tetrahedra)
			{
				Line l;

				if (t.V1 < t.V2) l = new Line(t.V1, t.V2);
				else l = new Line(t.V2, t.V1);
				if (!edgeHash.ContainsKey(l))
				{
					edgeHash.Add(l,edgeHash.Count);
				}

				if (t.V1 < t.V3) l = new Line(t.V1, t.V3);
				else l = new Line(t.V3, t.V1);
				if (!edgeHash.ContainsKey(l))
				{
					edgeHash.Add(l, edgeHash.Count);
				}

				if (t.V1 < t.V4) l = new Line(t.V1, t.V4);
				else l = new Line(t.V4, t.V1);
				if (!edgeHash.ContainsKey(l))
				{
					edgeHash.Add(l, edgeHash.Count);
				}

				if (t.V2 < t.V3) l = new Line(t.V2, t.V3);
				else l = new Line(t.V3, t.V2);
				if (!edgeHash.ContainsKey(l))
				{
					edgeHash.Add(l, edgeHash.Count);
				}

				if (t.V2 < t.V4) l = new Line(t.V2, t.V4);
				else l = new Line(t.V4, t.V2);
				if (!edgeHash.ContainsKey(l))
				{
					edgeHash.Add(l, edgeHash.Count);
				}

				if (t.V3 < t.V4) l = new Line(t.V3, t.V4);
				else l = new Line(t.V4, t.V3);
				if (!edgeHash.ContainsKey(l))
				{
					edgeHash.Add(l, edgeHash.Count);
				}
			}

			edges = new Line[edgeHash.Count];
			foreach (KeyValuePair<Line, int> pair in edgeHash)
			{
				edges[pair.Value] = pair.Key;
			}
		}

		private void fillStructures(UnstrGrid data)
		{
			tetrahedraList = new List<int>[divs[0], divs[1], divs[2], divs[3]];
			edgesList = new List<int>[divs[0], divs[1], divs[2], divs[3]];
			facesList = new List<int>[divs[0], divs[1], divs[2], divs[3]];
			pointsList = new List<int>[divs[0], divs[1], divs[2], divs[3]];

			for (int x = 0;x<divs[0];x++)
				for (int y = 0; y<divs[1];y++)
					for (int z = 0; z<divs[2];z++)
						for (int t = 0; t < divs[3]; t++)
						{
							tetrahedraList[x, y, z, t] = new List<int>();
							facesList[x, y, z, t] = new List<int>();
							edgesList[x, y, z, t] = new List<int>();
							pointsList[x, y, z, t] = new List<int>();
						}

			tetrahedra = data.GetCells(typeof(Tetrahedron))[0];
			points = data.Points;

			BoundingBox bb = data.BoundingBox;

			PointND sizes = (PointND)bb.MaxPoint - (PointND)bb.MinPoint;
			PointND minPoint = (PointND)bb.MinPoint - sizes * 0.02;
			sizes *= 1.04;
			steps = new double[4];
			mins = new double[4];
			for (int i = 0; i < 4; i++)
			{
				steps[i] = sizes[i]/divs[i];
				mins[i] = minPoint[i];
			}

			for (int i = 0; i < tetrahedra.Count;i++ )
			{
				addTetrahedron(i);
			}
			
			for(int i = 0; i<faces.Length;i++)
			{
				addTriangle(i);
			}

			for (int i = 0; i < edges.Length; i++)
			{
				addLine(i);
			}

			for (int i = 0; i< points.Count;i++)
			{
				addPoint(i);
			}
		}

		private bool cellMakesSense(int[] coords, int[] pointcoords, PointND point, double current)
		{
			double[] diff = new double[4];
			for (int i = 0; i < 4; i++)
			{
				if (coords[i] == pointcoords[i])
				{
					diff[i] = 0;
				}
				else if (coords[i] > pointcoords[i])
				{
					diff[i] = coords[i] * steps[i] + mins[i] - point[i];
				}
				else
				{
					diff[i] = point[i] - ((coords[i] + 1) * steps[i] + mins[i]);
				}
			}
			double minDistSq = diff[0] * diff[0] + diff[1] * diff[1] + diff[2] * diff[2] + diff[3] * diff[3];
			if ((current * current) > minDistSq) return (true);
			else return (false);
			//return (true);
		}		

		private double checkCell(int[] coords, PointND point, int pi)
		{
			double min = double.PositiveInfinity;
			double nd;
			
			foreach (int i in pointsList[coords[0], coords[1], coords[2], coords[3]])
			{
				PointND dif = (PointND)points[i] - point;
				double d = dif.dotProduct(dif);
				if (d < min) min = d;
			}

			foreach (int i in edgesList[coords[0], coords[1], coords[2], coords[3]])
			{
				if (lineChecks[i] == pi)
				{
					continue;
				}
				lineChecks[i] = pi;
				
				Line l = edges[i];

				if (PointToTetrahedronDistanceEvaluator.isProjectionOnEdge(point, out nd, (PointND)points[l.V1], (PointND)points[l.V2]))
				{
					if (nd<min) min = nd;
				}
			}			

			foreach (int i in facesList[coords[0], coords[1], coords[2], coords[3]])
			{
				if (faceChecks[i] == pi)
				{
					continue;
				}
				faceChecks[i] = pi;

				Triangle t = faces[i];

				if (PointToTetrahedronDistanceEvaluator.isProjectionOnFace(point, out nd, (PointND)points[t.V1], (PointND)points[t.V2], (PointND)points[t.V3]))
				{
					if (nd < min)
						min = nd;
				}
			}

			foreach (int i in tetrahedraList[coords[0], coords[1], coords[2], coords[3]])
			{
				if (tetraChecks[i] == pi)
				{
					continue;
				}
				tetraChecks[i] = pi;

				Tetrahedron t = (Tetrahedron)tetrahedra[i];

				if (PointToTetrahedronDistanceEvaluator.isProjectionOnTetrahedron(point, out nd, (PointND)points[t.V1], (PointND)points[t.V2], (PointND)points[t.V3], (PointND)points[t.V4]))
				{
					if (nd < min) min = nd;
				}
			}

			min = Math.Sqrt(min);

			return (min);
		}

		private double distanceTo(PointND point, int pi)
		{
			int[,] spans = new int[4, 2];

			int[] initCell = new int[4];

			int[] coords;

			// find cell where point lies
			spans[0, 0] = spans[0, 1] = initCell[0] = (int)Math.Floor((point[0] - mins[0]) / steps[0]);
			spans[1, 0] = spans[1, 1] = initCell[1] = (int)Math.Floor((point[1] - mins[1]) / steps[1]);
			spans[2, 0] = spans[2, 1] = initCell[2] = (int)Math.Floor((point[2] - mins[2]) / steps[2]);
			spans[3, 0] = spans[3, 1] = initCell[3] = (int)Math.Floor((point[3] - mins[3]) / steps[3]);
			for (int i = 0; i < 4; i++)
			{
				if (spans[i, 0] < 0) spans[i, 0] = 0;
				if (spans[i, 1] < 0) spans[i, 1] = 0;
				if (spans[i, 0] >= divs[i]) spans[i, 0] = divs[i] - 1;
				if (spans[i, 1] >= divs[i]) spans[i, 1] = divs[i] - 1;
			}

			double distance = double.PositiveInfinity;

			// check the cell
			double d = checkCell(new int[] { spans[0, 0], spans[1, 0], spans[2, 0], spans[3, 0] }, point, pi);
			if (d < distance) distance = d;

			bool working = true;

			while (working)
			{
				working = false;

				// try enlarge x
				int coord = 0;
				int nx = spans[coord, 1] + 1; // new x
				if ((nx < divs[coord]) && (distance > ((mins[coord] + nx * steps[coord]) - point[coord])))
				{
					// it makes sense, check all cells with nx
					for (int y = spans[1, 0]; y <= spans[1, 1]; y++)
						for (int z = spans[2, 0]; z <= spans[2, 1]; z++)
							for (int t = spans[3, 0]; t <= spans[3, 1]; t++)
							{
								coords = new int[] { nx, y, z, t };
								if (cellMakesSense(coords, initCell, point, distance))
								{
									d = checkCell(coords, point, pi);
									if (d < distance) distance = d;
								}
							}
					working = true;
					spans[coord, 1] = nx;
				}

				// try enlarge y
				coord = 1;
				int ny = spans[coord, 1] + 1; // new y
				if ((ny < divs[coord]) && (distance > ((mins[coord] + ny * steps[coord]) - point[coord])))
				{
					// it makes sense, check all cells with ny
					for (int x = spans[0, 0]; x <= spans[0, 1]; x++)
						for (int z = spans[2, 0]; z <= spans[2, 1]; z++)
							for (int t = spans[3, 0]; t <= spans[3, 1]; t++)
							{
								coords = new int[] { x, ny, z, t };
								if (cellMakesSense(coords, initCell, point, distance))
								{
									d = checkCell(coords, point, pi);
									if (d < distance) distance = d;
								}
							}
					working = true;
					spans[coord, 1] = ny;
				}

				// try enlarge z
				coord = 2;
				int nz = spans[coord, 1] + 1; // new z
				if ((nz < divs[coord]) && (distance > ((mins[coord] + nz * steps[coord]) - point[coord])))
				{
					// it makes sense, check all cells with nz
					for (int y = spans[1, 0]; y <= spans[1, 1]; y++)
						for (int x = spans[0, 0]; x <= spans[0, 1]; x++)
							for (int t = spans[3, 0]; t <= spans[3, 1]; t++)
							{
								coords = new int[] { x, y, nz, t };
								if (cellMakesSense(coords, initCell, point, distance))
								{
									d = checkCell(coords, point, pi);
									if (d < distance) distance = d;
								}
							}
					working = true;
					spans[coord, 1] = nz;
				}

				// try enlarge t
				coord = 3;
				int nt = spans[coord, 1] + 1; // new x
				if ((nt < divs[coord]) && (distance > ((mins[coord] + nt * steps[coord]) - point[coord])))
				{
					// it makes sense, check all cells with nx
					for (int y = spans[1, 0]; y <= spans[1, 1]; y++)
						for (int z = spans[2, 0]; z <= spans[2, 1]; z++)
							for (int x = spans[0, 0]; x <= spans[0, 1]; x++)
							{
								coords = new int[] { x, y, z, nt };
								if (cellMakesSense(coords, initCell, point, distance))
								{
									d = checkCell(coords, point, pi);
									if (d < distance) distance = d;
								}
							}
					working = true;
					spans[coord, 1] = nt;
				}

				// try shrink x
				coord = 0;
				nx = spans[coord, 0] - 1; // new x
				if ((nx >= 0) && (distance > (point[coord] - (mins[coord] + (nx + 1) * steps[coord]))))
				{
					// it makes sense, check all cells with nx
					for (int y = spans[1, 0]; y <= spans[1, 1]; y++)
						for (int z = spans[2, 0]; z <= spans[2, 1]; z++)
							for (int t = spans[3, 0]; t <= spans[3, 1]; t++)
							{
								coords = new int[] { nx, y, z, t };
								if (cellMakesSense(coords, initCell, point, distance))
								{
									d = checkCell(coords, point, pi);
									if (d < distance) distance = d;
								}
							}
					working = true;
					spans[coord, 0] = nx;
				}

				// try shrink y
				coord = 1;
				ny = spans[coord, 0] - 1; // new y
				if ((ny >= 0) && (distance > (point[coord] - (mins[coord] + (ny + 1) * steps[coord]))))
				{
					// it makes sense, check all cells with nx
					for (int x = spans[0, 0]; x <= spans[0, 1]; x++)
						for (int z = spans[2, 0]; z <= spans[2, 1]; z++)
							for (int t = spans[3, 0]; t <= spans[3, 1]; t++)
							{
								coords = new int[] { x, ny, z, t };
								if (cellMakesSense(coords, initCell, point, distance))
								{
									d = checkCell(coords, point, pi);
									if (d < distance) distance = d;
								}
							}
					working = true;
					spans[coord, 0] = ny;
				}

				// try shrink z
				coord = 2;
				nz = spans[coord, 0] - 1; // new z
				if ((nz >= 0) && (distance > (point[coord] - (mins[coord] + (nz + 1) * steps[coord]))))
				{
					// it makes sense, check all cells with nx
					for (int y = spans[1, 0]; y <= spans[1, 1]; y++)
						for (int x = spans[0, 0]; x <= spans[0, 1]; x++)
							for (int t = spans[3, 0]; t <= spans[3, 1]; t++)
							{
								coords = new int[] { x, y, nz, t };
								if (cellMakesSense(coords, initCell, point, distance))
								{
									d = checkCell(coords, point, pi);
									if (d < distance) distance = d;
								}
							}
					working = true;
					spans[coord, 0] = nz;
				}

				// try shrink t
				coord = 3;
				nt = spans[coord, 0] - 1; // new t
				if ((nt >= 0) && (distance > (point[coord] - (mins[coord] + (nt + 1) * steps[coord]))))
				{
					// it makes sense, check all cells with nx
					for (int y = spans[1, 0]; y <= spans[1, 1]; y++)
						for (int z = spans[2, 0]; z <= spans[2, 1]; z++)
							for (int x = spans[0, 0]; x <= spans[0, 1]; x++)
							{
								coords = new int[] { x, y, z, nt };
								if (cellMakesSense(coords, initCell, point, distance))
								{
									d = checkCell(coords, point, pi);
									if (d < distance) distance = d;
								}
							}
					working = true;
					spans[coord, 0] = nt;
				}
			}
			if (double.IsInfinity(distance))
			{
				throw new MveException(this, "Distance evaluated as infinity - probably there are no tetrahedra to test against.");
			}
			return (distance);
		}



		public override void Execute()
		{
			if (!IsAllDataSame() || (result == null))
			{
				UnstrGrid evaluated = (UnstrGrid)GetInput("Evaluated Mesh");
				UnstrGrid target = (UnstrGrid)GetInput("Target Mesh");

				divs = new int[] { 20, 20, 20, 20 };

				tetrahedra = target.GetCells(typeof(Tetrahedron))[0];
				points = target.Points;

				fillEdges(target);
				fillFaces(target);
				lineChecks = new int[edges.Length];
				for (int i = 0; i < lineChecks.Length; i++) lineChecks[i] = -1;
				faceChecks = new int[faces.Length];
				for (int i = 0; i < faceChecks.Length; i++) faceChecks[i] = -1;
				tetraChecks = new int[tetrahedra.Count];
				for (int i = 0; i < tetraChecks.Length; i++) tetraChecks[i] = -1;
				fillStructures(target);

				bool pnd = evaluated.Points.ElemType == typeof(PointND);

				UniformDataArray attr = new UniformDataArray(typeof(Scalar), evaluated.Points.Count);
				for (int i = 0; i < evaluated.Points.Count; i++)
				{
					
					Scalar s;
					if (pnd)
						s = new Scalar(distanceTo((PointND)evaluated.Points[i], i));
					else
					{
						Point3D p = (Point3D)evaluated.Points[i];
						s = new Scalar(distanceTo(new PointND(p.X, p.Y, p.Z, 0.0), i));
					}
					attr[i] = s;
					if ((i % 100) == 0) ProgressInfo((float)i / evaluated.Points.Count);
				}

				result = evaluated.ShallowCopy();
				result.AddPointAttrs("distances", attr);
				SetOutput("Result", result);
			}
			else
			{
				SetOutput("Result", result, true);
			}
		}
	}
}
