using System;
using Zcu.Mve.Core;
using Zcu.Mve.Numerics;

namespace Examples
{
  /// <summary>
  /// Module creates 100 samples of connected function in &lt;0, 10&gt; interval.
  /// Sample module with NonAutoUpdate and NonEvokeUpdate ports.
  /// </summary>
	[ModuleInfo("Milan Frank",
    "Module creates 100 samples of connected function in <0, 10> interval.",
     Category = "Examples")]
  [PortInfo("graph", "Output generated graph.")]
  [PortInfo("x", "Output indepenedent variable.")]
  [PortInfo("y", "Input dependent variable.")]
  public class GenerateGraph : Zcu.Mve.Core.Module
	{
		/// <summary>
		/// Creates ports.
		/// </summary>
    public GenerateGraph()
		{
      AddOutPort("graph", typeof(VectorND));
      AddOutPort("x", typeof(Scalar), false);
      AddInPort("y", typeof(Scalar), true, false);
		}
	
    /// <summary>
    /// Generates samples acording to dependent variable.
    /// </summary>
    public override void Execute()
    {
      VectorND vector = new VectorND(new double [100]);
      Scalar x = new Scalar();
      Scalar y;

      for (int i = 0; i < 100; i++)
      {
        x.Value = i * 0.1;
        SetOutput("x", x);
        // Execution of submap
        UpdateInput("y");
        y = (Scalar) GetInput("y");
        vector[i] = y.Value;
      }

      SetOutput("graph", vector);
    }
  }
}
