using System;
using System.Diagnostics;
using Zcu.Mve.Core;
using Zcu.Mve.Numerics;

namespace Examples
{
	/// <summary>
	/// Sample module. It sums two input numbers.
	/// </summary>
  [Zcu.Mve.Core.ModuleInfo("Milan Frank", "Sample module. It sums two input numbers.",
     Category = "Example",
     IconName = "Sumator.ico")]
  
  [PortInfo("A", "Input number A")]
  [PortInfo("B", "Input number B")]
  [PortInfo("Sum", "Sum of numbers A + B")]  
  public class Sumator : Zcu.Mve.Core.Module
	{
		private Scalar sum;

		/// <summary>
		/// Creates ports.
		/// </summary>
    public Sumator()
		{      
			AddInPort("A", typeof (Scalar));
			AddInPort("B", typeof (Scalar));
			AddOutPort("Sum", typeof (Scalar));

			sum = new Scalar();
    }

    /// <summary>
    /// Main method of any module. Called on demand of output data.
    /// </summary>
    public override void Execute()
    {
      // Get input data
      Scalar a = (Scalar) GetInput("A");
      Scalar b = (Scalar) GetInput("B");

      Console.WriteLine("Globals.useTrace = " + Globals.useTrace);
      // Check whether data are same as in previous Execute call.
      if (IsDataSame("A") && IsDataSame("B"))
      {      
        // In case all input data are same as in previous step we do not have to calculate them again.
        SetOutput("Sum", sum, true);
      }
      else
      {
        // Calculate sum.
        sum.Value = a.Value + b.Value;
        // Send data to outport.
        SetOutput("Sum", sum, false);
      }
    }
	}
}
